"""A simple program to monitor all peripherals found on the network."""

import os
import time
from pathlib import Path

import deimos
from deimos import LoopMethod, Overflow, LossOfContactPolicy

here = Path(__file__).parent.absolute()

testing = os.environ.get("DEIMOS_TESTING", False)

# Build a new control program to run at 5Hz (the minimum)
# in efficient (OS-scheduled) mode.
c = deimos.Controller("test2", str(here), 5.0)
c.loss_of_contact_policy = LossOfContactPolicy.reconnect_indefinite()
c.loop_method = LoopMethod.efficient()

# Scan for peripherals on the network
peripherals = c.scan()

# Add peripherals to the control program
print("Found peripherals on network:")
for i, p in enumerate(peripherals):
    print(f"  {type(p).__name__} SN {p.serial_number}")
    c.add_peripheral(f"p{i + 1}", p)

# Configure to write data to a CSV file
csv_dispatcher = deimos.dispatcher.CsvDispatcher(1, Overflow.wrap())
c.add_dispatcher("csv", csv_dispatcher)

# Add a calc that runs in-the-loop
five = deimos.calc.Constant(5.0, True)
c.add_calc("five", five)

# Run the control program nonblocking and poll latest values
if len(peripherals) > 0:
    h = c.run_nonblocking()
    try:
        for _ in range(20 if not testing else 1):  # poll for ~4 seconds
            time.sleep(0.2)
            snap = h.read()
            vals = list(snap.values.items())
            print(f"t={snap.timestamp} {snap.system_time} {vals[:3]}")
    finally:
        h.stop()
        try:
            h.join()
        except Exception as e:
            print(f"Run terminated with error: {e}")



"""
Processing chain for ITS-90 data:

           raw text    |     raw text converted to     | truncated data, w/ gaps  | symmetrized-ish     | table definitions for rust,
           format      |     bad csv and left-aligned  |   in actual CSV fmt      |                     | inverted via cubic interpolation
Website => type_k.tab => type_k_its90_csv_from_tab.csv => type_k_its90_trunc.csv => type_k_its90_mod.csv => its90*.txt
      manual        manual                           manual                  scripted                scripted

The first and last rows of the raw tables, which have incomplete data, are truncated.

For hot junction temperatures < 0C and > 0C, the cold junction error to the 0C-cold-junction
reading is mirrored about 0C cold junction temp in order to fill missing data with the best
available guess. At hot junction temp == 0C, both high and low cold junction temp table entries
are present, and are used directly.

This gives tables that are symmetric about 0C cold junction temperature everywhere _except_
at hot junction temp == 0C, where there is a very slight asymmetry present.

References
* [1] “ITS-90 Thermocouple Database, How to Use the Database.”
      Accessed: May 19, 2024. [Online]. Available:
      https://srdata.nist.gov/its90/useofdatabase/use_of_database.html#Coefficients%20Tables
"""

from pathlib import Path

import pandas as pd
import numpy as np
from interpn import MulticubicRectilinear

import matplotlib.pyplot as plt

here = Path(__file__).parent

df = pd.read_csv(here / "type_k_its90_trunc.csv", header=0, index_col=0)

# Symmetrize everything except hot junc temp == 0C
hjtemps_C = df.index.values
cjtemps_C = df.columns.values

for hjtemp_C in hjtemps_C:
    if int(hjtemp_C) < 0:
        # symmetrize low-temperature part
        for cjtemp_C in cjtemps_C:
            if int(cjtemp_C) < 0:
                cjerr = df.loc[hjtemp_C, cjtemp_C] - df.loc[hjtemp_C, "0"]
                df.loc[hjtemp_C, str(-int(cjtemp_C))] = df.loc[hjtemp_C, "0"] - cjerr
    elif int(hjtemp_C) == 0:
        pass  # Already populated
    elif int(hjtemp_C) > 0:
        # symmetrize high-temperature part
        for cjtemp_C in cjtemps_C:
            if int(cjtemp_C) > 0:
                cjerr = df.loc[hjtemp_C, cjtemp_C] - df.loc[hjtemp_C, "0"]
                df.loc[hjtemp_C, str(-int(cjtemp_C))] = df.loc[hjtemp_C, "0"] - cjerr
    else:
        raise NotImplementedError  # Just in case

# Save symmetrized-ish version for reference
df.to_csv(here / "type_k_its90_mod.csv")

# Generate rust table inputs
# ITS-90 cold junction temps are inverted,
# so we have to negate the celsius delta
# and then reverse the list so that it's sorted ascending
c_offs = 273.15  # conversion from degC to K
cjtemps = np.array(
    [(-float(x)) + c_offs for x in df.columns.values]
)[::-1]  # [K] from degC, cold junc. temp
hjtemps = np.array(
    [float(x) + c_offs for x in df.index.values]
)  # [K] from degC, hot junc. temp
voltages = df.values[:, ::-1] / 1e3  # [V] from mV, sensed voltages

# The tables map two temperatures to a voltage,
# but we need to map the two things we know (voltage and cold junction temp)
# to the one that we don't (hot junction temp) by inverting the tables.
#
# Since the voltage is monotonic in hot junction temperature for a given
# cold junction temperature, we can do this inversion by interpolating the
# hot junction temperatures on to a consistent voltage grid.

ngrid = 200
vmin = np.min(voltages)
vmax = np.max(voltages)
vgrid = np.linspace(
    vmin, vmax, ngrid
)  # Original tables have 163 hot junction temp points
hjtemps_interpolated = np.zeros((ngrid, len(cjtemps)))
for i, cjtemp in enumerate(cjtemps):
    v = voltages[:, i]
    assert len(v) == 163
    hjinterpolator = MulticubicRectilinear.new(
        [v], hjtemps, linearize_extrapolation=True
    )
    hjtemps_interpolated[:, i] = hjinterpolator.eval([vgrid])

# Save data to disk in little-endian binary format
with open(here / "its90_cold_junc_tables_rust.txt", "w") as f:
    step = cjtemps[1] - cjtemps[0]
    start = cjtemps[0]
    n = len(cjtemps)
    f.write('// [K] temperature at the connector "cold junction", regular grid\n')
    f.write(f"const COLD_JUNCTION_TEMP_START_K: f64 = {start};  // [K]\n")
    f.write(f"const COLD_JUNCTION_TEMP_STEP_K: f64 = {step};  // [K]\n")
    f.write(f"const COLD_JUNCTION_N: usize = {n};\n\n")

    step = vgrid[1] - vgrid[0]  # this is less float error than taking the mean diff
    start = vgrid[0]
    n = ngrid
    f.write(
        "// [V] sensed voltage at (or, hopefully, close to) the connector, regular grid\n"
    )
    f.write(f"const SENSED_VOLTAGE_START_V: f64 = {start}; // [V]\n")
    f.write(f"const SENSED_VOLTAGE_STEP_V: f64 = {step};  // [V]\n")
    f.write(f"const SENSED_VOLTAGE_N: usize = {n};\n\n")

    f.write(
        '// [K] Temperature at the probe ("hot junction"), 2d table mapping (sensed_voltage, cold_junction_temp) => hot_junction_temp.\n'
    )
    f.write("#[rustfmt::skip]\n")
    f.write(
        f"const HOT_JUNCTION_TEMP_K: [f64; {hjtemps_interpolated.size}] = "
        + str([float(x) for x in hjtemps_interpolated.flatten()])
        + ";  // [K]\n"
    )


plt.imshow(hjtemps_interpolated.T, aspect="equal", extent=[0, 1, 0, 1])
cs = plt.contour(hjtemps_interpolated.T, colors="k", extent=[0, 1, 0, 1])
plt.show()



from types import ModuleType
from typing import ClassVar, Protocol, Self, Literal

class CalcLike(Protocol):
    def to_json(self) -> str: ...

class DispatcherLike(Protocol):
    def to_json(self) -> str: ...

class PeripheralLike(Protocol):
    serial_number: int
    def to_json(self) -> str: ...

class SocketLike(Protocol):
    def to_json(self) -> str: ...

class Overflow:
    Wrap: ClassVar[Self]
    """Overwrite oldest data first."""

    NewFile: ClassVar[Self]
    """Create a new shard."""

    Error: ClassVar[Self]
    """Emit an error and shut down."""

    @staticmethod
    def wrap() -> Wrap:
        """Wrap back to the beginning of the file."""
        ...
    @staticmethod
    def new_file() -> NewFile:
        """Create a new file on overflow."""
        ...
    @staticmethod
    def error() -> Error:
        """Error on overflow."""
        ...

class LoopMethod:
    Performant: ClassVar[Self]
    """
    Use 100% of a CPU to protect timing.
    This increases maximum usable control frequency.
    """

    Efficient: ClassVar[Self]
    """
    Use operating system scheduling to reduce CPU usage
    at the expense of degraded cycle performance.
    Typically viable up to about 50Hz control rate.
    """

    @staticmethod
    def performant() -> Performant:
        """Use 100% of a CPU to protect timing."""
        ...
    @staticmethod
    def efficient() -> Efficient:
        """Use operating system scheduling for lower CPU usage."""
        ...

class Termination:
    Timeout: ClassVar[Self]
    """End the control program after some duration from the start of the first cycle."""

    Scheduled: ClassVar[Self]
    """End the control program at a specific UTC system time."""

    @staticmethod
    def timeout_s(s: float) -> Timeout:
        """End after `s` seconds from the start of the first cycle."""
        ...
    @staticmethod
    def scheduled_epoch_ns(ns: int) -> Scheduled:
        """End at a specified absolute system time in UTC nanoseconds."""
        ...

class LossOfContactPolicy:
    Terminate: ClassVar[Self]
    """Terminate the control program."""

    Reconnect: ClassVar[Self]
    """Attempt to reconnect to the peripheral"""

    @staticmethod
    def terminate() -> Terminate:
        """Construct a policy that terminates the control program."""
        ...
    @staticmethod
    def reconnect_s(timeout_s: float) -> Reconnect:
        """Construct a reconnect policy with a timeout in seconds."""
        ...
    @staticmethod
    def reconnect_indefinite() -> Reconnect:
        """Construct a reconnect policy with no timeout."""
        ...

class Snapshot:
    @property
    def system_time(self) -> str: ...
    """End-of-cycle UTC system time in RFC3339 format with nanoseconds."""
    @property
    def timestamp(self) -> int: ...
    """End-of-cycle time in monotonic nanoseconds since start of control program."""
    @property
    def values(self) -> dict[str, float]: ...
    """Latest values by channel name."""

class RunHandle:
    def stop(self) -> None:
        """Signal the controller to stop."""
        ...
    def is_running(self) -> bool:
        """Check if the controller thread is still running."""
        ...
    def is_ready(self) -> bool:
        """Check if the controller has completed its first cycle."""
        ...
    def join(self) -> str:
        """Wait for the controller thread to finish."""
        ...
    def latest_row(self) -> tuple[str, int, list[float]]:
        """Get the latest row: (system_time, timestamp, channel_values)."""
        ...
    def headers(self) -> list[str]:
        """Column headers including timestamp/time."""
        ...
    def read(self) -> Snapshot:
        """Read the latest row mapped to header names."""
        ...
    def available_inputs(self) -> list[str]:
        """List peripheral inputs that can be written manually."""
        ...
    def write(self, values: dict[str, float]) -> None:
        """Write values to peripheral inputs not driven by calcs."""
        ...

class Controller:
    """
    The control program that communicates with hardware peripherals,
    runs calculations, and dispatches data.
    """

    def __init__(self, op_name: str, op_dir: str, rate_hz: float) -> None:
        """Build a new controller.

        `rate_hz` will be rounded to the nearest nanosecond when converted
        to the sample period.

        This constructor does not run the controller or attach any peripherals.
        """
        ...
    def run(self) -> str:
        """Run the control program."""
        ...
    def run_nonblocking(
        self,
        latest_value_cutoff_freq: float | None = None,
        wait_for_ready: bool = True,
    ) -> RunHandle:
        """Run the control program on a separate thread and return a handle
        for coordination.

        Args:
            latest_value_cutoff_freq: Optional second-order Butterworth low-pass filter
                                      cutoff frequency to apply to latest-value data.
                                      If the selected frequency is outside the viable
                                      range for the filter, the cutoff frequency will
                                      be clamped to the viable bounds and a warning
                                      will be emitted.
            wait_for_ready: Block until the controller has completed its first cycle.
        """
        ...
    def scan(self, timeout_ms: int = 10) -> list[PeripheralLike]:
        """Scan the local network (and any other attached sockets)
        for available peripherals."""
        ...
    def available_inputs(self) -> list[str]:
        """List peripheral inputs that can be written manually."""
        ...
    def add_peripheral(self, name: str, p: PeripheralLike) -> None:
        """Register a peripheral with the control program"""
        ...
    def attach_hootl_driver(
        self,
        peripheral_name: str,
        transport: HootlTransport,
        end_epoch_ns: int | None = None,
    ) -> HootlRunHandle:
        """Wrap an existing peripheral with a hootl wrapper and start its driver."""
        ...
    def add_calc(self, name: str, calc: CalcLike) -> None:
        """Add a calc to the expression graph that runs on every cycle"""
        ...
    def add_dataframe_dispatcher(
        self,
        name: str,
        max_size_megabytes: int,
        overflow_behavior: Overflow,
    ) -> dispatcher.DataFrameHandle:
        """Add an in-memory dataframe dispatcher and return its shared handle."""
        ...
    def add_dispatcher(self, name: str, dispatcher: DispatcherLike) -> None:
        """Add a dispatcher via a JSON-serializable dispatcher instance."""
        ...
    def dispatcher_names(self) -> list[str]: ...
    def remove_dispatcher(self, name: str) -> bool: ...
    def add_socket(self, name: str, socket: SocketLike) -> None:
        """Add a socket via a JSON-serializable socket instance."""
        ...
    def remove_socket(self, name: str) -> bool:
        """Remove a socket by name."""
        ...
    def set_peripheral_input_source(self, input_field: str, source_field: str) -> None:
        """Connect an entry in the calc graph to a
        command to be sent to the peripheral."""
        ...
    def clear_calcs(self) -> None:
        """Remove all calcs."""
        ...
    def clear_peripherals(self) -> None:
        """Remove all peripherals."""
        ...
    def clear_dispatchers(self) -> None:
        """Remove all dispatchers."""
        ...
    def clear_sockets(self) -> None:
        """Remove all sockets."""
        ...

    @property
    def op_name(self) -> str:
        """
        The name of the operation.
        Used to set database table names, set log and data file names, etc.
        """
        ...
    @op_name.setter
    def op_name(self, v: str) -> None: ...
    @property
    def op_dir(self) -> str:
        """
        The directory where this operation's logs and other data will be placed,
        and where calcs with linked configuration (like a SequenceMachine) can
        find their linked files or folders by relative path.
        """
        ...
    @op_dir.setter
    def op_dir(self, v: str) -> None: ...
    @property
    def dt_ns(self) -> int:
        """[ns] control program cycle period."""
        ...
    @dt_ns.setter
    def dt_ns(self, v: int) -> None: ...
    @property
    def rate_hz(self) -> float:
        """[Hz] control program cycle rate."""
        ...
    @rate_hz.setter
    def rate_hz(self, v: float) -> None: ...
    @property
    def peripheral_loss_of_contact_limit(self) -> int:
        """Number of missed packets from the controller that indicates disconnection."""
        ...
    @peripheral_loss_of_contact_limit.setter
    def peripheral_loss_of_contact_limit(self, v: int) -> None: ...
    @property
    def controller_loss_of_contact_limit(self) -> int:
        """Number of missed packets from a peripheral that indicates disconnection."""
        ...
    @controller_loss_of_contact_limit.setter
    def controller_loss_of_contact_limit(self, v: int) -> None: ...
    @property
    def termination_criteria(self) -> Termination | None:
        """Criteria for exiting the control program."""
        ...
    @termination_criteria.setter
    def termination_criteria(self, v: Termination | None) -> None: ...
    @property
    def loss_of_contact_policy(self) -> LossOfContactPolicy:
        """
        The response of the control program when a peripheral disconnects during run.
        """
        ...

    @loss_of_contact_policy.setter
    def loss_of_contact_policy(self, v: LossOfContactPolicy) -> None: ...
    @property
    def loop_method(self) -> LoopMethod:
        """
        The loop waiting method for the controller.

        Busywaiting is performant, but inefficient;
        relying on the operating system for scheduling is efficient, but not performant.
        """
        ...
    @loop_method.setter
    def loop_method(self, v: LoopMethod) -> None: ...
    @property
    def enable_manual_inputs(self) -> bool:
        """Whether manual input overrides should be applied during the control loop."""
        ...
    @enable_manual_inputs.setter
    def enable_manual_inputs(self, v: bool) -> None: ...

class _CalcBase:
    def to_json(self) -> str:
        """Serialize to typetagged JSON so Python can pass into trait handoff."""
        ...
    @classmethod
    def from_json(cls, s: str) -> Self:
        """Deserialize from typetagged JSON."""
        ...

class _DispatcherBase:
    def to_json(self) -> str:
        """Serialize to typetagged JSON so Python can pass into trait handoff."""
        ...
    @classmethod
    def from_json(cls, s: str) -> Self:
        """Deserialize from typetagged JSON."""
        ...

class _PeripheralBase:
    serial_number: int
    def to_json(self) -> str:
        """Serialize to typetagged JSON so Python can pass into trait handoff."""
        ...
    @classmethod
    def from_json(cls, s: str) -> Self:
        """Deserialize from typetagged JSON."""
        ...

class _SocketBase:
    def to_json(self) -> str:
        """Serialize to typetagged JSON so Python can pass into trait handoff."""
        ...
    @classmethod
    def from_json(cls, s: str) -> Self:
        """Deserialize from typetagged JSON."""
        ...

InterpMethod = Literal["linear", "left", "right", "nearest"]
Time = list[float]
Value = list[float]
Name = str
Sequence = dict[Name, tuple[Time, Value, InterpMethod]]
TimeoutTargetState = str

class _CalcModule(ModuleType):
    class Affine(_CalcBase):
        """A slope and offset, y = ax + b."""
        def __init__(
            self,
            input_name: str,
            slope: float,
            offset: float,
            save_outputs: bool,
        ) -> None: ...

    class Butter2(_CalcBase):
        """Single-input, single-output Butterworth low-pass
        filter implemented with `flaw::butter2`."""
        def __init__(
            self, input_name: str, cutoff_hz: float, save_outputs: bool
        ) -> None: ...

    class Constant(_CalcBase):
        """Simplest calc that does anything at all."""
        def __init__(self, y: float, save_outputs: bool) -> None: ...

    class InverseAffine(_CalcBase):
        """Derive input voltage from linear amplifier reading.

        First subtracts the output offset, then divides by the slope.
        """
        def __init__(
            self,
            input_name: str,
            slope: float,
            offset: float,
            save_outputs: bool,
        ) -> None: ...

    class Pid(_CalcBase):
        """A PID controller with simple saturation for anti-windup."""
        def __init__(
            self,
            measurement_name: str,
            setpoint_name: str,
            kp: float,
            ki: float,
            kd: float,
            max_integral: float,
            save_outputs: bool,
        ) -> None: ...

    class Polynomial(_CalcBase):
        """Polynomial calibration: y = c0 + c1*x + c2*x^2 + ...

        with an attached note that should include traceability info
        like a sensor serial number.
        Coefficients ordered by increasing polynomial order.
        """
        def __init__(
            self,
            input_name: str,
            coefficients: list[float],
            note: str,
            save_outputs: bool,
        ) -> None: ...

    class RtdPt100(_CalcBase):
        """Derive input voltage from amplifier output."""
        def __init__(self, resistance_name: str, save_outputs: bool) -> None: ...

    class Sin(_CalcBase):
        """Sin wave between `low` and `high` with a period of `period_s`
        and phase offset of `offset_s`."""
        def __init__(
            self,
            period_s: float,
            offset_s: float,
            low: float,
            high: float,
            save_outputs: bool,
        ) -> None: ...

    class TcKtype(_CalcBase):
        """Calculate a K-type thermocouple's temperature reading in (K) from voltage,
        using the ITS-90 method for cold-junction correction.
        """
        def __init__(
            self,
            voltage_name: str,
            cold_junction_temperature_name: str,
            save_outputs: bool,
        ) -> None: ...

    class SequenceMachineInner(_CalcBase):
        """A lookup-table sequence machine that follows a set procedure during
        each sequence, and transitions between sequences based on set criteria.

        Unlike most calcs, the names of the inputs and outputs of this calc
        are not known at compile-time, and are assembled from inputs instead.
        """
        def __init__(self, entry: str) -> None: ...
        @staticmethod
        def load_folder(path: str) -> Self:
            """Read a configuration json and sequence CSV files from a folder.

            The folder must contain one json representing a [MachineCfg] and
            some number of CSV files each representing a [Sequence].
            """
            ...
        def save_folder(self, path: str) -> None:
            """Save a configuration json and sequence CSV files to a folder."""
            ...
        def get_entry(self) -> str:
            """The sequence that is the entrypoint for the machine."""
            ...

        def set_entry(self, entry: str) -> None:
            """Set the sequence that is the entrypoint for the machine."""
            ...

        def get_link_folder(self) -> str | None:
            """
            The folder, by relative path from `op_dir`,
            that this machine will automatically reload
            configuration from during init, if any.
            """
            ...

        def set_link_folder(self, link_folder: str | None) -> None:
            """
            Set the folder, by relative path from `op_dir`,
            that this machine will automatically reload
            configuration from during init, if any.
            """
            ...

        def get_timeout(self, sequence: str) -> TimeoutTargetState | None:
            """
            The state that this sequence will transition to when it times out.
            If None, this sequence will loop.
            """
            ...

        def set_timeout(self, sequence: str, target: TimeoutTargetState | None) -> None:
            """
            Set the state that this sequence will transition to when it times out.
            If None, this sequence will loop.
            """
            ...

        def add_sequence(
            self,
            name: str,
            tables: Sequence,
            timeout: TimeoutTargetState | None,
        ) -> None:
            """Add a sequence state to the machine."""
            ...

        def add_constant_thresh_transition(
            self,
            source_target: tuple[str, str],
            channel: str,
            op: tuple[str, float],
            threshold: float,
        ) -> None:
            """Add a constant threshold transition for a sequence."""
            ...

        def add_channel_thresh_transition(
            self,
            source_target: tuple[str, str],
            channel: str,
            op: tuple[str, float],
            threshold_channel: str,
        ) -> None:
            """Add a channel threshold transition for a sequence."""
            ...

        def add_lookup_thresh_transition(
            self,
            source_target: tuple[str, str],
            channel: str,
            op: tuple[str, float],
            threshold_lookup: tuple[list[float], list[float], str],
        ) -> None:
            """Add a lookup threshold transition for a sequence."""
            ...

calc: _CalcModule

class HootlTransport:
    @staticmethod
    def thread_channel(name: str) -> Self:
        """A thread channel with this name."""
        ...

    @staticmethod
    def unix_socket(name: str) -> Self:
        """A unix socket with this name."""
        ...

    @staticmethod
    def udp() -> Self:
        """UDP transport bound to PERIPHERAL_RX_PORT (one hootl driver at a time)."""
        ...

class HootlRunHandle:
    def stop(self) -> None: ...
    def is_running(self) -> bool: ...
    def join(self) -> None: ...
    def __enter__(self) -> Self: ...
    def __exit__(self, exc_type: object, exc: object, tb: object) -> bool: ...

class _PeripheralModule(ModuleType):
    class AnalogIRev2(_PeripheralBase):
        def __init__(self, serial_number: int) -> None: ...

    class AnalogIRev3(_PeripheralBase):
        def __init__(self, serial_number: int) -> None: ...

    class AnalogIRev4(_PeripheralBase):
        def __init__(self, serial_number: int) -> None: ...

    class DeimosDaqRev5(_PeripheralBase):
        def __init__(self, serial_number: int) -> None: ...

    class DeimosDaqRev6(_PeripheralBase):
        def __init__(self, serial_number: int) -> None: ...

    class HootlPeripheral(_PeripheralBase):
        """Peripheral wrapper that emits mock outputs using driver-owned state.

        Note: attach this via Controller.attach_hootl_driver to keep the shared
        driver state intact; JSON roundtrips reset the link.
        """
        def __init__(self, inner: PeripheralLike) -> None: ...

    HootlTransport = HootlTransport

    HootlRunHandle = HootlRunHandle

    class HootlDriver:
        def __init__(
            self,
            inner: PeripheralLike,
            transport: HootlTransport,
            end_epoch_ns: int | None = None,
        ) -> None: ...
        """A way to operate a hootl driver from outside the control program."""

        def run_with(self, controller: Controller) -> HootlRunHandle: ...
        """Start the driver attached to this controller."""

peripheral: _PeripheralModule

class _SocketModule(ModuleType):
    class UnixSocket(_SocketBase):
        """Implementation of Socket trait for stdlib unix socket."""
        def __init__(self, name: str) -> None: ...

    class UdpSocket(_SocketBase):
        """Implementation of Socket trait for stdlib UDP socket on IPV4."""
        def __init__(self) -> None: ...

    class ThreadChannelSocket(_SocketBase):
        """Socket implementation that consumes a user channel of the same name.
        Only one peripheral can be connected per thread socket."""
        def __init__(self, name: str) -> None: ...

socket: _SocketModule

class _DispatcherModule(ModuleType):
    class CsvDispatcher(_DispatcherBase):
        """A plain-text CSV data target, which uses a pre-sized file
        to prevent sudden increases in write latency during file resizing.

        On reaching the end of the configured data size, it can be configured to
        * Wrap (and start overwriting the existing data from the beginning),
        * Start a new file, or
        * Error, logging that the file is full and closing comms with the controller.

        depending on the appropriate response for the task at hand.

        Each line in this CSV format is fixed-width, meaning the line length will not
        change with each line. As a result, it is possible to read to a specific time
        or line in O(1) time by simple arithmetic rather than by reading the whole file.
        This guarantee will be broken when crossing from year 9999 to year 10000
        and so on, or if non-finite values are encountered in measurements, calcs,
        or metrics.

        Writes to disk on a separate thread to avoid blocking the control loop.
        """
        def __init__(
            self,
            chunk_size_megabytes: int,
            overflow_behavior: Overflow,
        ) -> None: ...

    class TimescaleDbDispatcher(_DispatcherBase):
        """Either reuse or create a new table in a TimescaleDB postgres
        database and write to that table.

        Depending on the buffer time window, this dispatcher will either use a simple
        INSERT query to send one row of values at a time, or use COPY INTO syntax
        to send buffered batches of rows all at once.

        Performs database transactions on on a separate thread to avoid blocking the
        control loop.

        Does not support TLS, and as a result, is recommended for communication with
        databases on the same internal network, not on the open web.
        """
        def __init__(
            self,
            dbname: str,
            host: str,
            user: str,
            token_name: str,
            buffer_time_ns: int,
            retention_time_hours: int,
        ) -> None: ...

    class DataFrameDispatcher(_DispatcherBase):
        """
        WARNING: This dispatcher requires a direct link in the backend, and can't be
        initialized properly from Python. Instead, it can be initialized by
        `Controller.add_dataframe_dispatcher`.

        Store collected data in in-memory columns, moving columns into
        a dataframe behind a shared reference at termination.

        To avoid deadlocks, the dataframe is not updated until after
        the run is complete. The dataframe is cleared at the start of each run.
        """
        def __init__(
            self,
            max_size_megabytes: int,
            overflow_behavior: Overflow,
        ) -> None: ...
        def handle(self) -> dispatcher.DataFrameHandle: ...

    class DataFrameHandle:
        """Shared handle for reading collected dataframe data.

        Rows are returned in their stored order. When the dispatcher uses wrap
        overflow, older rows are overwritten in-place, so ordering may not be
        strictly chronological. Use `time()` and `timestamp()` for time columns;
        `columns()` returns channel values only.
        """
        def columns(self) -> dict[str, list[float]]: ...
        def time(self) -> list[str]: ...
        def timestamp(self) -> list[int]: ...

    class LatestValueDispatcher(_DispatcherBase):
        """
        WARNING: This dispatcher requires a direct link in the backend, and can't be
        initialized properly from Python.

        Dispatcher that always keeps the latest row available via a shared handle.

        The latest row is updated on a dedicated worker thread so the controller
        loop does not block on reader access. Before the first dispatch, the row
        is initialized with NaN channel values and a UNIX_EPOCH timestamp.
        """
        def __init__(self) -> None: ...

    class ChannelFilter(_DispatcherBase):
        """Wraps another dispatcher and forwards a subset of channels by name.

        Channel names are validated during initialization. Missing or duplicate
        names are treated as errors, and the output order follows `channels`.
        Timestamps are forwarded unchanged; only per-channel values are filtered.
        """
        def __init__(
            self,
            inner: DispatcherLike,
            channels: list[str],
        ) -> None: ...

    class DecimationDispatcher(_DispatcherBase):
        """Wraps another dispatcher and forwards every Nth row.

        The first row is forwarded, then `nth - 1` rows are dropped, repeating.
        Timestamps and channel values are passed through unchanged for forwarded rows.
        """
        def __init__(
            self,
            inner: DispatcherLike,
            nth: int,
        ) -> None: ...

    class LowPassDispatcher(_DispatcherBase):
        """Wraps another dispatcher and applies a 2nd order Butterworth low-pass
        filter per channel.

        The cutoff frequency is clamped to a stable ratio of the sample rate.
        The filter is primed on the first row by setting each channel's steady-state
        value to the incoming sample.
        """
        def __init__(
            self,
            inner: DispatcherLike,
            cutoff_hz: float,
        ) -> None: ...

dispatcher: _DispatcherModule



"""Run examples to make sure they don't go stale"""

import os
import pathlib
import runpy

import pytest

os.environ["DEIMOS_TESTING"] = "True"

EXAMPLES_DIR = pathlib.Path(__file__).parent / "../examples"
EXAMPLES = [
    EXAMPLES_DIR / x
    for x in os.listdir(EXAMPLES_DIR)
    if (os.path.isfile(EXAMPLES_DIR / x) and x.endswith(".py"))
]


@pytest.mark.parametrize("fp", EXAMPLES)
def test_example(fp: pathlib.Path):
    try:
        runpy.run_path(str(fp), run_name="__main__")
    except:
        print(f"Failed to run example {fp}")
        raise



from importlib.metadata import PackageNotFoundError, version

from .deimos import (
    calc,
    dispatcher,
    peripheral,
    socket,
    Controller,
    Overflow,
    Termination,
    LossOfContactPolicy,
    LoopMethod,
)

try:
    __version__ = version("deimos-daq")
except PackageNotFoundError:
    __version__ = "0.0.0"

__all__ = [
    "calc",
    "dispatcher",
    "peripheral",
    "socket",
    "Controller",
    "Overflow",
    "Termination",
    "LossOfContactPolicy",
    "LoopMethod",
    "__version__",
]



"""Test markdown code blocks"""

import pathlib
from mktestdocs import check_md_file


def test_docs():
    # check_md_file(fpath=pathlib.Path("docs") / "index.md")
    check_md_file(fpath=pathlib.Path(".") / "README.md")



"""Smoke test for HOOTL mockups with full peripheral/socket/dispatcher coverage."""

from __future__ import annotations
from math import isnan

import tempfile
import time
from contextlib import ExitStack
from pathlib import Path

from deimos import (
    Controller,
    LoopMethod,
    Overflow,
    Termination,
    dispatcher,
    peripheral,
    socket,
)

THREAD_CHANNEL1 = "hootl_thread1"
THREAD_CHANNEL2 = "hootl_thread2"
UNIX_SOCKET = "ctrl"
RATE_HZ = 20.0
RUN_TIMEOUT_S = 10.0  # Accommodate slow CI runners (we don't actually run this long)
LATEST_FILTER_HZ = 5.0
HAS_UNIX_SOCKET = hasattr(socket, "UnixSocket") and hasattr(
    peripheral.HootlTransport, "unix_socket"
)


def _metric_channels(peripheral_name: str) -> list[str]:
    """Limited channel list to reduce disk I/O during testing"""
    return [
        f"{peripheral_name}.metrics.cycle_time_ns",
        f"{peripheral_name}.metrics.loss_of_contact_counter",
    ]


def _make_transport(kind: str, name: str | None) -> peripheral.HootlTransport:
    if kind == "thread":
        if name is None:
            raise ValueError("thread transport requires a name")
        return peripheral.HootlTransport.thread_channel(name)
    if kind == "unix":
        if not HAS_UNIX_SOCKET:
            raise RuntimeError(
                "unix socket transport is not available on this platform"
            )
        if name is None:
            raise ValueError("unix transport requires a name")
        return peripheral.HootlTransport.unix_socket(name)
    if kind == "udp":
        return peripheral.HootlTransport.udp()
    raise ValueError(f"Unknown transport kind: {kind}")


def _build_controller(
    op_dir: Path, loop_method: LoopMethod
) -> tuple[Controller, list[tuple[str, peripheral.HootlTransport]]]:
    ctrl = Controller(op_name="smoketest", op_dir=str(op_dir), rate_hz=RATE_HZ)
    ctrl.termination_criteria = Termination.timeout_s(RUN_TIMEOUT_S)
    ctrl.loop_method = loop_method

    ctrl.clear_sockets()
    ctrl.add_socket("thread1", socket.ThreadChannelSocket(THREAD_CHANNEL1))
    ctrl.add_socket("thread2", socket.ThreadChannelSocket(THREAD_CHANNEL2))
    if HAS_UNIX_SOCKET:
        ctrl.add_socket("unix", socket.UnixSocket(UNIX_SOCKET))
    ctrl.add_socket("udp", socket.UdpSocket())

    ctrl.add_dispatcher("csv", dispatcher.CsvDispatcher(1, Overflow.wrap()))
    ctrl.add_dispatcher("latest_value", dispatcher.LatestValueDispatcher())
    ctrl.add_dispatcher(
        "channel_filter",
        dispatcher.ChannelFilter(
            dispatcher.LatestValueDispatcher(),
            _metric_channels("analog_rev2"),
        ),
    )
    ctrl.add_dispatcher(
        "decimation",
        dispatcher.DecimationDispatcher(dispatcher.LatestValueDispatcher(), 2),
    )
    ctrl.add_dispatcher(
        "low_pass",
        dispatcher.LowPassDispatcher(
            dispatcher.LatestValueDispatcher(), LATEST_FILTER_HZ
        ),
    )
    ctrl.add_dataframe_dispatcher("dataframe", 1, Overflow.wrap())

    specs = [
        ("analog_rev2", peripheral.AnalogIRev2, 1001, ("thread", THREAD_CHANNEL1)),
        ("analog_rev3", peripheral.AnalogIRev3, 1002, ("thread", THREAD_CHANNEL2)),
    ]
    if HAS_UNIX_SOCKET:
        specs.extend(
            [
                ("analog_rev4", peripheral.AnalogIRev4, 1003, ("unix", "per_analog4")),
                ("daq_rev5", peripheral.DeimosDaqRev5, 1004, ("unix", "per_daq5")),
            ]
        )
    specs.append(("daq_rev6", peripheral.DeimosDaqRev6, 1005, ("udp", None)))

    attachments: list[tuple[str, peripheral.HootlTransport]] = []
    for name, cls, serial, (transport_kind, transport_name) in specs:
        ctrl.add_peripheral(name, cls(serial))
        transport = _make_transport(transport_kind, transport_name)
        attachments.append((name, transport))

    return ctrl, attachments


def _run_controller(
    loop_method: LoopMethod,
    blocking: bool,
    latest_value_cutoff: float | None,
) -> None:
    with tempfile.TemporaryDirectory(prefix="deimos-smoketest-") as tmp_dir:
        op_dir = Path(tmp_dir)
        ctrl, attachments = _build_controller(op_dir, loop_method)

        with ExitStack() as stack:
            for name, transport in attachments:
                stack.enter_context(ctrl.attach_hootl_driver(name, transport))

            if blocking:
                # Use shorter timeout for blocking mode, since we don't need to read
                # values live.
                ctrl.termination_criteria = Termination.timeout_s(0.2)
                ctrl.run()
                return

            handle = ctrl.run_nonblocking(latest_value_cutoff)
            try:
                time.sleep(0.2)  # Accommodate slow CI runners
                snapshot = handle.read()
                expected = _metric_channels("analog_rev2")
                for channel in expected:
                    assert channel in snapshot.values
                
                for k, v in snapshot.values.items():
                    assert not isnan(v), f"Channel {k} did not read successfully."
                
            finally:
                if handle.is_running():
                    print("Sending controller stop signal from Python.")
                    handle.stop()
                handle.join()

    # Pause for hootl drivers to shut down
    time.sleep(0.1)


def test_hootl_smoketest() -> None:
    for loop_method in [LoopMethod.performant(), LoopMethod.efficient()]:
        _run_controller(loop_method, blocking=True, latest_value_cutoff=None)
        _run_controller(loop_method, blocking=False, latest_value_cutoff=None)
        _run_controller(
            loop_method, blocking=False, latest_value_cutoff=LATEST_FILTER_HZ
        )



"""
A showcase of drivers for imitating hardware from software
to test the software's interface with hardware.
"""

import time
from contextlib import ExitStack
from pathlib import Path
from deimos import Controller, peripheral, socket, Termination, LoopMethod

HAS_UNIX_SOCKET = hasattr(socket, "UnixSocket") and hasattr(
    peripheral.HootlTransport, "unix_socket"
)


def main() -> None:
    here = Path(__file__).parent.resolve()

    for loop_method in [LoopMethod.performant(), LoopMethod.efficient()]:
        print(f"Testing with loop method {loop_method}")

        # Set up HOOTL peripherals (drivers are attached via the controller)
        ctrl = Controller(op_name="mockup_demo", op_dir=str(here / "op"), rate_hz=20.0)
        ctrl.termination_criteria = Termination.timeout_s(2.0)
        ctrl.loop_method = loop_method

        ctrl.clear_sockets()
        ctrl.add_socket("mockup_chan", socket.ThreadChannelSocket("mockup_chan"))
        if HAS_UNIX_SOCKET:
            ctrl.add_socket("ctrl", socket.UnixSocket("ctrl"))
        ctrl.add_socket("udp", socket.UdpSocket())  # Included by default, but cleared

        ctrl.add_peripheral("mock_thread", peripheral.DeimosDaqRev6(1))
        if HAS_UNIX_SOCKET:
            ctrl.add_peripheral("mock_unix", peripheral.DeimosDaqRev6(2))
        ctrl.add_peripheral("mock_udp", peripheral.DeimosDaqRev6(3))

        # Run
        with ExitStack() as stack:
            # Run the peripheral mockups, which will wait for the controller
            # to send a request to bind
            stack.enter_context(
                ctrl.attach_hootl_driver(
                    "mock_thread",
                    peripheral.HootlTransport.thread_channel("mockup_chan"),
                )
            )
            if HAS_UNIX_SOCKET:
                stack.enter_context(
                    ctrl.attach_hootl_driver(
                        "mock_unix",
                        peripheral.HootlTransport.unix_socket("mockup_unix"),
                    )
                )
            stack.enter_context(
                ctrl.attach_hootl_driver(
                    "mock_udp",
                    peripheral.HootlTransport.udp(),
                )
            )

            # Get list of inputs available to set manually.
            # This is also available from the RunHandle during operation.
            manual_inputs = ctrl.available_inputs()
            print("Manual inputs available:")
            for name in manual_inputs[:3]:  # Adjust slicing to list more input names
                print(f"    {name}")
            print(f"    ...and {len(manual_inputs) - 3} more")

            # Run the controller, which will bind the peripheral mockups
            start = time.perf_counter()
            handle = ctrl.run_nonblocking()

            try:
                time.sleep(0.2)  # Accommodate slow CI runners
                handle.write({"mock_thread.dac0": 0.0})

                # Make sure we had stable communication with all the peripheral mockups
                for k, v in handle.read().values.items():
                    if "loss_of_contact_counter" in k:
                        # When testing on macos runners, we're not able to set
                        # core affinity, which causes sporadic packet loss.
                        assert v < 2.0, f"Missed packet: {k} = {v:.0f}"
            except Exception:
                end = time.perf_counter()
                print(
                    "Sending termination signal to run handle"
                    f" from Python after {end - start:.2f}s."
                )
                handle.stop()
                raise
            finally:
                handle.join()


if __name__ == "__main__":
    main()



"""
Processing chain for DIN-43-760 / IEC-751 data:
                                                           table definitions for rust,
           pdf table       |       copy to CSV         |   with corrected misprint,
                           |   and correct misprint    |   interpolated on to regular resistance grid
Website => "RTD table.pdf" => pt100_rtd_din-43-760.csv => pt100_rtd_din-43-76_rust.txt
      manual             manual                     scripted

The resistance values at 706C and 798 degC are misprinted in the PDF I used, which is apparent from
the fact that their values are not monotonically increasing and do not match the `delta` column.
They have been corrected in the CSV document using the `delta` column to calculate the value
at those temperatures from the value at the previous temperature and the `delta` column.

References
* DIN-43-760 (IEC-751) PT100 RTD tables, acquired as PDF
 from https://web.mst.edu/cottrell/ME240/Resources/Temperature/RTD%20table.pdf
"""

from pathlib import Path

import pandas as pd
import numpy as np
from interpn import MulticubicRectilinear

import matplotlib.pyplot as plt

here = Path(__file__).parent

df = pd.read_csv(here / "pt100_rtd_din-43-760.csv", header=0, index_col=0)

c_offs = 273.15  # conversion from degC to K

temps_orig = df.index.values.astype(np.float64) + c_offs  # [K] from [degC]
resistances_orig = df["Ohms"].values  # [ohm]

# Interpolate on to regular resistance grid so that the in-the-loop interpolation is fast
#   Use more grid points in interpolated version to maintain accuracy
#   across regions with changing spacing in the original data
ngrid = 2 * len(temps_orig)

temp_interpolator = MulticubicRectilinear.new([resistances_orig], temps_orig, linearize_extrapolation=True)
resistances = np.linspace(np.min(resistances_orig), np.max(resistances_orig), ngrid, endpoint=True)  # [ohm]
temps = temp_interpolator.eval([resistances])  # [K]
temps = [float(x) for x in temps]

# Save data to disk in little-endian binary format
with open(here / "pt100_rtd_din-43-76_rust.txt", "w") as f:
    step = float(resistances[1] - resistances[0])
    start = float(resistances[0])
    n = len(resistances)
    f.write('// [ohm] probe resistance, regular grid\n')
    f.write(f"const PROBE_RESISTANCE_START_OHM: f64 = {start};  // [ohm]\n")
    f.write(f"const PROBE_RESISTANCE_STEP_OHM: f64 = {step};  // [ohm]\n")
    f.write(f"const PROBE_RESISTANCE_N: usize = {n};\n\n")

    f.write(
        '// [K] Temperature at the probe, 1d table mapping resistance => probe_temp.\n'
    )
    f.write("#[rustfmt::skip]\n")
    f.write(
        f"const PROBE_TEMP_K: [f64; {len(temps)}] = "
        + str(temps)
        + ";  // [K]\n"
    )

plt.plot(resistances, temps, color='k', linewidth=2)
plt.xlabel("resistance [ohm]")
plt.ylabel("temperature [K]")
plt.show()
